"""
Reactor pattern to call queued functions in a separate thread.

Functions can be queued to be called by the reactor main thread, or by a thread
in the reactor's thread pool.  Functions can be be queued to be called
immediately (as soon as calling thread is available), or queued to be called at
a specified later time.

"""
__author__ = "Andrew Gillis"
__license__ = "http://www.opensource.org/licenses/mit-license.php"
__maintainer__ = "Andrew Gillis"
__email__ = "gillis.andrewj@gmail.com"

import threading
import traceback
import heapq
import time
from collections import Callable
try:
    import queue
except ImportError:
    import Queue as queue


class Completion(object):

    """
    Completion represents the result of an asynchronous computation.

    A Completion object is returned by a Reactor's call_() methods.  It holds a
    value that may become available at some point.  The Completion class is
    equivalent to the Future class offered by other concurrency frameworks.

    The Completion object allows the caller to wait for and retrieve the result
    of a function called by the Reactor.  This not only includes any return
    from the called function, but any exception as well.

    A Completion object is also used to identify a scheduled function to
    cancel, and it can be used to check if a scheduled call has been canceled.

    """

    slots = ('_call_done', '_rsp', '_exception', '_traceback_str', '_task_id',
             '_result_q')
    def __init__(self, result_q):
        self._call_done = threading.Event()
        self._rsp = None
        self._exception = None
        self._traceback_str = None

        # For canceling scheduled tasks.
        self._task_id = None

        # Optional queue to put this Completion on when completed.
        self._result_q = result_q

    def result(self, timeout=None):
        """Return the value returned by the call.

        If the call has not yet completed then wait up to timeout seconds or
        until call complete if timeout is None.  It timeout occurs, then raise
        a RuntimeError exception.  If canceled before completion, then a
        ThreadError exception is raised.

        If the call raised an exception, then this methods returns the same
        exception.

        """
        if self._call_done.wait(timeout):
            if self._exception:
                raise self._exception

            return self._rsp

        if self.canceled():
            # Canceled before completed.
            raise threading.ThreadError('call canceled before completion')

        # Timeout waiting for completion.
        raise RuntimeError('timeout waiting for result')

    def done(self):
        """Return True if call completed or canceled."""
        return self._call_done.is_set()

    def traceback(self):
        """Return the traceback string from exception in call."""
        return self._traceback_str

    def canceled(self):
        return self._task_id is False


class _Deferred(object):
    # Internal object used to transition from execution in main thread to
    # execution in pool thread.
    #
    slots = ('action_tuple',)
    def __init__(self, func, args=None, callback=None):
        self.action_tuple = (func, args, callback)


class Reactor(threading.Thread):

    """
    The Reactor executes queued functions in a single thread.  Functions can be
    executed immediately, in the order they are enqueued, or they can be
    executed at some later time.  Functions may also be executed, immediately
    or later, in threads separate from the reactor's main thread.

    All functions executed by the reactor have an associated Completion object.
    A Completion object allows the caller to wait for and retrieve the result
    of a function called by the reactor.

    The reason to use a Reactor is to execute functions in a separate thread,
    asynchronously, from the thread of the caller.

    """

    def __init__(self, thread_pool=None):
        """Initialize Reactor thread object and internal thread pool (if used).

        Arguments:
        thread_pool -- ThreadPool object to use to submit tasks to.  If this is
                       None, then the reactor will not use a thread pool and
                       will start threads directly.  Use a thread pool to
                       throttle concurrency and to avoid thread creation
                       overhead.

        """
        threading.Thread.__init__(self, None, None, None)
        self._call_queue = queue.Queue()
        self.idle_wake_sec = None
        if thread_pool is not None:
            self._thread_pool = thread_pool
        else:
            self._thread_pool = None

        self._task_heap = []

    def call(self, func, args=None, callback=None, result_q=None):
        """Enqueue a function for the reactor to run in its main thread.

        The caller can wait on the response using the returned Completion
        object. Running a function in the reactor's main thread means that no
        other functions will be running in the reactor's main thread at the
        same time.  This can be used in place of other synchronization.

        Arguments:
        func     -- Function to execute.
        args     -- Argument tuple to pass to function.
        callback -- Optional callback that takes return from func as argument.
        result_q -- Optional response queue to place finished Completion on.

        Return:
        Completion object.

        """
        assert(self.is_alive()), 'reactor is not started'
        assert(isinstance(func, Callable)), 'function is not callable'
        assert(callback is None or
               isinstance(callback, Callable)),'callback is not callable'
        assert(result_q is None or
               isinstance(result_q, queue.Queue)), 'result_q is not Queue'
        future = Completion(result_q)
        self._call_queue.put((func, args, future, callback, None, None))
        return future

    def call_later(self, delay, func, args=None, callback=None, result_q=None):
        """Run a function in the reactor's main thread at a later time.

        The caller can wait on the response using the returned Completion
        object.

        Arguments:
        delay    -- Seconds remaining until call give as int or float.
        func     -- Function to execute.
        args     -- Argument tuple to pass to function.
        callback -- Optional callback that takes return from func as argument.
        result_q -- Optional response queue to place complete Result on.

        Return:
        Completion object.

        """
        assert(self.is_alive()), 'reactor is not started'
        assert(isinstance(func, Callable)), 'function is not callable'
        assert(callback is None or
               isinstance(callback, Callable)), 'callback is not callable'
        if delay:
            when = time.time() + delay
        else:
            when = 0

        return self.call_at(when, func, args, callback, result_q)

    def call_at(self, when, func, args=None, callback=None, result_q=None):
        """Run a function in the reactor's main thread at the specified time.

        Arguments:
        when     -- Absolute timestamp specifying when to call func, given as
                    int or float, using the same reference as time().
        func     -- Function to execute.
        args     -- Argument tuple to pass to function.
        callback -- Optional callback that takes return from func as argument.
        result_q -- Optional response queue to place complete Result on.

        Return:
        Completion object.

        """
        assert(self.is_alive()), 'reactor is not started'
        assert(isinstance(func, Callable)), 'function is not callable'
        assert(callback is None or
               isinstance(callback, Callable)), 'callback is not callable'
        future = Completion(result_q)
        self._call_queue.put((func, args, future, callback, when, None))
        return future

    def call_in_thread(self, func, args=None, callback=None, result_q=None):
        """
        Run the given function in a separate thread.

        The caller can wait on the response using the returned Completion
        object.  A thread from the reactor's internal thread pool is used to
        avoid the overhead of creating a new thread.

        Arguments:
        func     -- Function to execute.
        args     -- Argument tuple to pass to function.
        callback -- Optional callback that takes return from func as argument.
        result_q -- Optional response queue to place finished Completion on.

        Return:
        Completion object.

        """
        assert(self.is_alive()), 'reactor is not started'
        assert(isinstance(func, Callable)), 'function is not callable'
        assert(callback is None or
               isinstance(callback, Callable)), 'callback is not callable'
        future = Completion(result_q)
        args = (func, args, future, callback)
        if self._thread_pool:
            self._thread_pool.submit(self._thread_wrapper, args)
        else:
            th = threading.Thread(target=self._thread_wrapper, args=args)
            th.daemon = True
            th.start()
        return future

    def call_in_thread_later(self, time_until_call, func, args=None,
                             callback=None, result_q=None):
        """Run the given function in a separate thread, at a later time.

        The caller can wait on the response using the returned Completion
        object.  A thread from the reactor's internal thread pool is used to
        avoid the overhead of creating a new thread.

        Arguments:
        time_until_call -- Seconds remaining until call.
        func            -- Function to execute.
        args            -- Argument tuple to pass to function.
        callback        -- Optional callback that takes return from func as
                           argument.
        result_q        -- Optional response queue to place Completion on.

        Return:
        Completion object.

        """
        assert(self.is_alive()), 'reactor is not started'
        assert(isinstance(func, Callable)), 'function is not callable'
        assert(callback is None or
               isinstance(callback, Callable)), 'callback is not callable'
        if time_until_call:
            action_time = time.time() + time_until_call
        else:
            action_time = 0

        future = Completion(result_q)
        self._call_queue.put((func, args, future, callback, action_time, True))
        return future

    def cancel_scheduled(self, future):
        """Cancel a call scheduled to be called at a later time.

        Arguments:
        future -- Completion object returned from call_later() or from
                  call_in_thread_later().

        Return:
        True if scheduled call was successfully canceled.  False if scheduled
        call was not found (already executed) or already canceled.

        """
        # The Completion's task_id cannot be checked in this method.  The
        # request must be placed on the reactor's call queue to ensure that a
        # scheduled task request is received by the reactor before this request
        # to cancel it is received.
        tmp_future = self.call(self._cancel_scheduled, future)
        # Wait for reactor to remove task, and return whether or not the task
        # was removed.
        return tmp_future.result()

    def shutdown(self, shutdown_threadpool=True):
        """Cause the reactor thread to exit gracefully.

        Arguments:
        shutdown_threadpool -- If True, and Reactor is using a ThreadPool, then
                               shutdown the ThreadPool.  Set this to false if
                               the ThreadPool instance is still in use outside
                               of this Reactor.

        Return:
        True if shutdown, False if Reactor not started.

        """
        if not self.is_alive():
            #print "thread is not started"
            return False

        self._call_queue.put(None)
        # Wait for thread to exit.
        self.join()

        # Shutdown threadpool after reactor main thread has exited to ensure no
        # more tasks are submitted to the thread pool.
        if self._thread_pool is not None and shutdown_threadpool:
            self._thread_pool.shutdown()

        return True

    def qsize(self):
        """Return number of items in the message queue."""
        return self._call_queue.qsize()

    def empty(self):
        """Return True is no items in the message queue."""
        return self._call_queue.empty()

    def defer_to_thread(self, func, args=None, callback=None):
        """
        Defer execution from main thread to pooled thread, and use the caller's
        Completion object to return the result of the additional execution.

        Arguments:
        func     -- Function to execute in pooled thread.
        args     -- Arguments to pass to func.
        callback -- Optional callback that takes return from func as argument.

        Returns:
        Special object recognized by Reactor to set up deferred call.

        """
        return _Deferred(func, args, callback)

    #==========================================================================
    # Private methods follow:
    #==========================================================================

    def _thread_wrapper(self, func, args, future, callback):
        try:
            if args is None:
                rsp = func()
            elif isinstance(args, (tuple, list, set)):
                rsp = func(*args)
            elif isinstance(args, dict):
                rsp = func(**args)
            else:
                rsp = func(args)
        except Exception as e:
            rsp = None
            future._exception = e
            future._traceback_str = traceback.format_exc()

        if isinstance(rsp, _Deferred):
            func, args, new_callback = rsp.action_tuple
            if new_callback is not None:
                callback = new_callback
            if self._thread_pool:
                self._thread_pool.submit(
                    self._thread_wrapper, (func, args, future, callback))
            else:
                th = threading.Thread(
                    target=self._thread_wrapper,
                    args=(func, args, future, callback))
                th.daemon = True
                th.start()
            return

        if callback:
            try:
                callback(rsp)
            except Exception as e:
                rsp = None
                future._exception = e
                future._traceback_str = traceback.format_exc()

        # Set call done only after callback has completed.  This is necessary
        # for code waiting on a Completion to be able to reliably see an
        # exception from a callback.
        future._rsp = rsp
        future._call_done.set()

        if future._result_q:
            result_q = future._result_q
            future._result_q = None
            try:
                result_q.put(future)
            except queue.Full:
                pass

    def _process_scheduled_queue(self):
        """Process all scheduled tasks that are currently due.

        """
        task_heap = self._task_heap
        # Execute the first task without checking the time, because this method
        # is only called when a scheduled task is due.
        task_time, task_id, task_data = heapq.heappop(task_heap)

        # If this task has been removed, then do nothing.
        if task_data is None:
            return

        def do_sched_task(func, args, future, callback, call_at, in_thread):
            if in_thread:
                if self._thread_pool is not None:
                    # Call thread pool to execute scheduled function.
                    self._thread_pool.submit(
                        self._thread_wrapper, (func, args, future, callback))
                else:
                    # Create new thread to execute scheduled function.
                    th = threading.Thread(
                        target=self._thread_wrapper,
                        args=(func, args, future, callback))
                    th.daemon = True
                    th.start()
            else:
                #print '===> calling scheduled function'
                self._thread_wrapper(func, args, future, callback)

        do_sched_task(*task_data)
        if task_heap:
            now = time.time()
            next_time = task_heap[0][0]
            while task_heap and (not next_time or next_time <= now):
                task_time, task_id, task_data = heapq.heappop(task_heap)
                do_sched_task(*task_data)
                if task_heap:
                    next_time = task_heap[0][0]

    def _cancel_scheduled(self, future):
        """Mark a scheduled call, in scheduled task heap, as canceled.

        """
        task_id = future._task_id
        if task_id is None or task_id is False:
            # Task is not scheduled or is already canceled.
            return False
        task_heap = self._task_heap
        for idx, heap_item in enumerate(task_heap):
            task_time, heap_task_id, task_data = heap_item
            if heap_task_id == task_id:
                func, args, future, callback, call_at, in_thread = task_data
                future._rsp = None
                future._task_id = False
                future._call_done.set()
                # Need to keep task_id to preserve heap invariant.
                task_heap[idx] = (task_time, heap_task_id, None)
                return True
        return False

    def run(self):
        """Function run in thread, started by the start() method.

        """
        call_queue = self._call_queue
        task_heap = self._task_heap
        next_task_id = 0
        while True:
            sleep_time = None
            # If there are items on the scheduled task heap, then sleep until
            # next scheduled task is due.
            if task_heap:
                now = time.time()
                # Look at timer at top of heap.
                action_time = task_heap[0][0]
                if not action_time or action_time <= now:
                    sleep_time = 0
                else:
                    sleep_time = action_time - now

            try:
                #print '===> sleeping for %s seconds' % (sleep_time,)
                # Get the next function from the call queue.
                work_item = call_queue.get(True, sleep_time)
                if work_item is None:
                    break
                func, args, future, callback, call_at, in_thread = work_item

                # If this is a new scheduled task.
                if call_at:
                    future._task_id = next_task_id
                    heapq.heappush(task_heap,
                                   (call_at, next_task_id, work_item))
                    next_task_id += 1
                else:
                    self._thread_wrapper(func, args, future, callback)

            except queue.Empty:
                # Woke up because it is time do next scheduled task.
                #print '===> woke up to do next scheduled task'
                self._process_scheduled_queue()
                continue

            except Exception:
                print('ERROR: '+str(traceback.format_exc()))

        return True

